"""
product initialization stuff
"""
import os
import featuremonkey
from .composer import get_composer

_product_selected = False

def select_product():
    """
    binds the frozen context the selected features
    
    should be called only once - calls after the first call have
    no effect
    """
    global _product_selected
    if _product_selected:
        #tss already bound ... ignore
        return

    _product_selected = True
    from django_productline import context

    featuremonkey.add_import_guard('django.conf')
    featuremonkey.add_import_guard('django.db')
    os.environ['DJANGO_SETTINGS_MODULE'] = 'django_productline.settings'
    contextfile = os.environ['PRODUCT_CONTEXT_FILENAME']
    equationfile = os.environ['PRODUCT_EQUATION_FILENAME']
    #bind context and compose features
    context.bind_context(contextfile)
    get_composer().select_equation(equationfile)
    featuremonkey.remove_import_guard('django.conf')
    featuremonkey.remove_import_guard('django.db')

def get_wsgi_application():
    """
    returns the wsgi application for the selected product
    
    this function is called by featuredjango.wsgi to get the wsgi
    application object
    
    if you need to refine the wsgi application object e.g. to add
    wsgi middleware please refine django.core.wsgi.get_wsgi_application directly.
    """
    #make sure the product is selected before importing and constructing wsgi app
    select_product()
    #return (possibly refined) wsgi application
    from django.core.wsgi import get_wsgi_application
    return get_wsgi_application()



#refinement for django_productline.settings

def refine_INSTALLED_APPS(original):
    return original + ['django.contrib.staticfiles']

# Absolute path to the directory static files should be collected to.
# Don't put anything in this directory yourself; store your static files
# in apps' "static/" subdirectories and in STATICFILES_DIRS.
# Example: "/home/media/media.lawrence.com/static/"
from django_productline.context import PRODUCT_CONTEXT
introduce_STATIC_ROOT = PRODUCT_CONTEXT.DATA_DIR + '/generated_static/'

# URL prefix for static files.
# Example: "http://media.lawrence.com/static/"
introduce_STATIC_URL = '/static/'

# Additional locations of static files
introduce_STATICFILES_DIRS = [
    # Put strings here, like "/home/html/static" or "C:/www/django/static".
    # Always use forward slashes, even on Windows.
    # Don't forget to use absolute paths, not relative paths.
]

# List of finder classes that know how to find static files in
# various locations.
introduce_STATICFILES_FINDERS = [
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
#    'django.contrib.staticfiles.finders.DefaultStorageFinder',
]



"""
django-productline root urlconf

urlpatterns are constructed by refining django_productline.urls.get_urls.

Here, get_urls is called to get the (composed) urlpatterns.
Django uses these to construct the root RegexUrlResolver.
"""
from django_productline import urls

urlpatterns = urls.get_urls()



import featuremonkey
from featuremonkey import helpers

class Composer(featuremonkey.Composer):
    """
    The django_productline composer is based on
    featuremonkey and adds a new composition rule
    that is useful for django product lines.

    Django makes extensive use of metaprogramming to
    provide the db.models, forms etc.
    Unfortunately, this renders the introduction mechanism of
    featuremonkey useless in some cases, e.g. when trying
    to introduce new fields in a django model.

    However, django provides the ``contribute_to_class`` hook
    that can be used to circumvent such problems in some cases.

    Therefore, this composer additionally provides a ``contribute``
    operation, that works a lot like ``introduce`` but uses
    ``contribute_to_class`` instead of ``setattr`` underneath.

    Contributions may not be specified as factory functions.
    Use explicit attribute assignment.
    """

    def _contribute(self, role, target_attrname, transformation, base):
        if hasattr(base, target_attrname):
            raise featuremonkey.CompositionError(
                'Cannot contribute "%s" of "%s" to "%s"!'
                ' Attribute exists already!' % (
                    target_attrname,
                    helpers._get_role_name(role),
                    helpers._get_base_name(base),
                )
            )
        if callable(transformation):
            raise featuremonkey.CompositionError(
                'Cannot contribute "%s" of "%s" to "%s"!'
                ' Contributions must not be callable!' % (
                    target_attrname,
                    helpers._get_role_name(role),
                    helpers._get_base_name(base),
                )
            )

        transformation.contribute_to_class(base, target_attrname)

    def _apply_transformation(self, role, base, transformation, attrname):
        if attrname.startswith('contribute_'):
            target_attrname = attrname[len('contribute_'):]
            self._contribute(role, target_attrname, transformation, base)
        else:
            super(Composer, self)._apply_transformation(
                role, base, transformation, attrname
            )

def get_composer():
    return Composer()



"""
WSGI config for featuredjango project.

This module contains the WSGI application used by Django's development server
and any production WSGI deployments. It should expose a module-level variable
named ``application``. Django's ``runserver`` and ``runfcgi`` commands discover
this application via the ``WSGI_APPLICATION`` setting.

Usually you will have the standard Django WSGI application here, but it also
might make sense to replace the whole Django WSGI application with a custom one
that later delegates to the Django one. For example, you could introduce WSGI
middleware here, or combine a Django application with an application of another
framework.
For products built with django-productline this can be done by refining
django.core.wsgi.get_wsgi_application directly.
"""
from django_productline.startup import get_wsgi_application
application = get_wsgi_application()



"""
the product context captures environment and configuration settings
that are specific to each product, e.g. each product requires a different
database configuration.

Use the context only for very specific stuff that NEEDS to 
be configured on a product basis.

The context is loaded from a file in json format.

"""
import json

PRODUCT_CONTEXT = None

class ContextAccessor(object):
    """
    provides nice interface to access the product context
    
    only reading is allowed! Don`t write to the context, please!
    """

    def __init__(self, data):
        """
        :param: data dict to wrap
        """
        self._data = data

    def __getattr__(self, name):
        """
        makes uppercase keys of wrapped dict available using
        dot notation
        """
        if name.isupper():
            try:
                return self._data[name]
            except KeyError:
                pass

        raise AttributeError


def bind_context(context_filename):
    """
    loads context from file and binds to it
    
    :param: context_filename absolute path of the context file
    
    called by featuredjango.startup.select_product
    prior to selecting the individual features
    """
    global PRODUCT_CONTEXT
    if PRODUCT_CONTEXT is None:
        with open(context_filename) as contextfile:
            context = json.loads(contextfile.read())
            context['PRODUCT_CONTEXT_FILENAME'] = context_filename
            PRODUCT_CONTEXT = ContextAccessor(context)
    else:
        #bind_context called but context already bound
        #harmless rebind (with same file) is ignored
        #otherwise this is a serious error
        if PRODUCT_CONTEXT.PRODUCT_CONTEXT_FILENAME != context_filename:
            raise Exception('product context bound multiple times using different data!')




def refine_get_urls(original):
    def get_urls():
        from django.conf.urls import patterns, include
        from django.contrib import admin

        admin.autodiscover()

        return original() + patterns('',
            (r'^admin/', include(admin.site.urls)),
        )
    return get_urls



from ape import tasks
from decorator import decorator
import os

@tasks.register_helper
@decorator#preserves signature of wrappee
def requires_product_environment(func, *args, **kws):
    """
    task decorator that makes sure that the product environment
    of django_productline is activated:
    -context is bound
    -features have been composed
    """
    from django_productline import startup
    startup.select_product()
    return func(*args, **kws)

@tasks.register
@tasks.requires_product_environment
def manage(*args):
    """
    call django management tasks
    """
    from django.core.management import execute_from_command_line
    execute_from_command_line(['ape manage'] + list(args))


@tasks.register
def prepare_db():
    """
    create the database
    """
    #the base feature assumes sqlite
    #or that the database is managed manually
    #other features can easily add support for
    #a specific database server and creation mechanism
    #by refining this task
    pass


@tasks.register
def prepare_db_schema():
    """
    create the database schema
    """
    tasks.manage('syncdb', '--noinput')


@tasks.register
@tasks.requires_product_environment
def prepare():
    """
    prepare product for deployment

    does stuff like:
    -setting up the db and schema
    -generate webserver config
    ... everything to get the web application setup

    must be rerun after feature selection and/or the
    product context is changed
    """
    from django_productline.context import PRODUCT_CONTEXT
    if not os.path.isdir(PRODUCT_CONTEXT.DATA_DIR):
        os.mkdir(PRODUCT_CONTEXT.DATA_DIR)
    tasks.prepare_db()
    tasks.prepare_db_schema()

@tasks.register
def dev():
    """
    run the development server
    """
    tasks.manage('runserver')

@tasks.register
def deploy():
    """
    deploy application
    """
    #other features may add support for mod_wsgi, uwsgi, gunicorn,...
    #by _replacing_ this method
    tasks.dev()



from ape import tasks

@tasks.register
@tasks.requires_product_environment
def prepare_staticfiles(force=False):
    """
    collect static files for production httpd

    If run with ``settings.DEBUG==True``, this is a no-op
    unless ``force`` is set to ``True``
    """
    #noise reduction: only collectstatic if not in debug mode
    from django.conf import settings
    if force or not settings.DEBUG:
        tasks.manage('collectstatic')


def refine_prepare(original):
    """
    adds call to build_staticfiles
    """
    def prepare():
        original()
        tasks.prepare_staticfiles()
    return prepare



#! /usr/bin/env python
import os
import django_productline
from setuptools import setup, find_packages

def read(fname):
    try:
        return open(os.path.join(os.path.dirname(__file__), fname)).read()
    except IOError:
        return ''

setup(
    name='django-productline',
    version='0.2',
    description='an approach to feature-oriented software development(FOSD) of web application product lines based on django',
    long_description=read('README.rst'),
    license='The MIT License',
    keywords='django, FOSD, FOP, feature-oriented-programming, product-line, web-application',
    author='Hendrik Speidel',
    author_email='hendrik@schnapptack.de',
    url="https://github.com/henzk/django-productline",
    packages=find_packages(),
    package_dir={'django_productline': 'django_productline'},
    package_data={'django_productline': []},
    include_package_data=True,
    scripts=[],
    zip_safe=False,
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent'
    ],
    install_requires=[
        'Django',
        'decorator',
        'django-overextends',
        'south',
        'featuremonkey>=0.2.2',
        'ape>=0.2.0'
    ]
)




def refine_get_urls(original):
    """
    serve static files
    
    in production the webserver should serve requested
    static files itself and never let requests to /static/*
    get to the django application.
    """
    def get_urls():
        from django.contrib.staticfiles.urls import staticfiles_urlpatterns
        return original() + staticfiles_urlpatterns()
    return get_urls



def select(composer):
    from . import settings
    import django_productline.settings
    composer.compose(settings, django_productline.settings)

    from . import urls
    import django_productline.urls
    composer.compose(urls, django_productline.urls)



"""

``django_productline.urls`` exports the function ``get_urls``.
It is called through ``django_productline.root_urlconf`` which is registered as django`s ``ROOT_URLCONF``.

To introduce new urlpatterns, features may refine ``get_urls``.
The convention is to specify the refinement in a module called ``urls`` within the feature module.

Example::

    #in myfeature.feature.select()
    from . import urls #import the refinement definition
    import featuredjango.urls #import the base module
    #apply the refinement to the base module
    featuremonkey.compose(urls, featuredjango.urls)

    #in myfeature.urls
    def refine_get_urls(original):
        def get_urls():
            '''
            introduce myfeature`s views
            '''
            from django.urls import patterns
            return original() + patterns('',
                (r'^foo/$', 'myfeature.views.foo'),
                (r'^bar/(\d{4})/$', 'myfeature.views.bar'),
            )

        return get_urls

"""

def get_urls():
    """
    refine this to add urlpatterns
    """
    return []
